/*
 * farm_mandelbrot.cpp
 *
 *      Author: Steffen Ernsting <s.ernsting@uni-muenster.de>
 *
 * -------------------------------------------------------------------------------
 *
 * The MIT License
 *
 * Copyright 2014 Steffen Ernsting <s.ernsting@uni-muenster.de>,
 *                Herbert Kuchen <kuchen@uni-muenster.de.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *
 */

#include <iostream>
#include <cstdlib>
#include <fstream>

#include "mpi.h"
#include "farm.h"
#include "pipe.h"
#include "initial.h"
#include "final.h"
#include "functors.h"

// aux
int iCounter = 0;
int jCounter = 0;
int** mandelbrotSet;

struct Result
{
public:
	int row, col, res;

	MSL_USERFUNC
	Result()
			: row(0), col(0), res(0)
	{
	}

	MSL_USERFUNC
	Result(int r, int c, int re)
			: row(r), col(c), res(re)
	{
	}
};

std::ostream& operator<<(std::ostream& os, const Result& obj)
{
	os << "(" << obj.row << ", " << obj.col << ", " << obj.res << ")";
	return os;
}

// Argument functor for the atomic skeleton.
// Iterates the mandelbrot sequence (z_n+1 -> (z_n)^2 + c, z_0 = 0).
struct Iterate : msl::FarmFunctor<int, Result>
{
  int maxIters, zoom, rows, cols;
  float maxAbs = 4.0f, center_x, center_y;
  float dx, dy, l, r, t, b;

  Iterate(int mIters, float cx, float cy, int z, int rs, int cs)
    : maxIters(mIters), zoom(z), rows(rs), cols(cs), center_x(cx), center_y(cy)
  {
    l = center_x - ((float) cols / (zoom * 2));
    r = center_x + ((float) cols / (zoom * 2));
    t = center_y - ((float) rows / (zoom * 2));
    b = center_y + ((float) rows / (zoom * 2));

    dx = (r - l) / cols;
    dy = (b - t) / rows;
  }

	MSL_USERFUNC
	Result operator()(int n) const
	{
    int iters = 0;

    int row = n / rows;
    int col = n % rows;
    float real = l + col*dx; float tmpReal = real;
    float imag = t + row*dy; float tmpImag = imag;
    float nextReal;

    while ((real * real) + (imag * imag) <= maxAbs && iters < maxIters) {
      nextReal = (real * real) - (imag * imag) + tmpReal;
      imag = (2 * real * imag) + tmpImag;
      real = nextReal;
      ++iters;
    }

		return Result(row, col, iters);
	}
};

int iBlock = 0;
int jBlock = 0;
// Argument functor for the Initial skeleton.
// Calculates complex numbers depending on the resolution. Returning a null 
// pointer indicates that all tasks have been created. A STOP signal will be 
// sent to all processes in order to stop the calculation. 
struct Init
{
  int rows, cols, block_size;

  Init(int r, int c, int bs)
    : rows(r), cols(c), block_size(bs)
  {}

	int* operator()()
	{
    int index = (iCounter+(iBlock*block_size))*cols + jCounter+(jBlock*block_size);
    if (index == rows*cols) {
      return 0;
    }
    int* res = new int(index);

    if (jCounter == block_size-1) {
      jCounter = 0;
      if (iCounter == block_size-1) {
        iCounter = 0;
        if (jBlock == cols/block_size-1) {
          iBlock++;
          jBlock = 0;
        } else {
          jBlock++;
        }
      } else {
        iCounter++;
      }
    } else {
      jCounter++;
    }

//    if (iCounter == rows * cols) {
//      return 0;
//    }
//
//    int* res = new int(iCounter);
//    iCounter++;

    return res;
	}
};

void reset()
{
  iBlock = jBlock = iCounter = jCounter = 0;
}

// Argument functor for the Final skeleton.
int resCount;
struct Fin
{
	void operator()(Result result)
	{
		mandelbrotSet[result.row][result.col] = result.res;
	}
};

void printResult(int rows, int cols, int maxIters)
{
	// Set up the output file
	std::ostringstream headerStream;
	headerStream << "P6\n";
	headerStream << cols << ' ' << rows << '\n';
	headerStream << "255\n";
	std::ofstream fileStream("mandelbrot.ppm", std::ios::out | std::ios::binary);
	fileStream << headerStream.str();

  unsigned char r, g, b;
  int iters;

	for (int y = 0; y < rows; y++) {
		for (int x = 0; x < cols; ++x) {
		  iters = mandelbrotSet[y][x];
			if (iters < maxIters) {
        r = ((iters & 63) << 2);
        g = ((iters << 3) & 255);
        b = ((iters >> 8) & 255);
      } else {
        r = 0;
        g = 0;
        b = 0;
      }

			fileStream << r << g << b;
		}
	}

	// Tidy up (probably unnecessary)
	fileStream.flush();
	fileStream.close();
}

// run as: mpirun -np "2+#workers" mandel #workers
int main(int argc, char** argv)
{
	bool debugCommunication = 0;
	msl::initSkeletons(argc, argv, debugCommunication);

  int nWorkers = 1;
  if (msl::Muesli::num_total_procs < 3) {
    if (msl::isRootProcess()) {
      std::cout << "Start with at least 3 processes!" << std::endl;
    }
    msl::fail_exit();
  } else {
    nWorkers = msl::Muesli::num_total_procs - 2;
  }

  int rows = 1024, cols = 1024, nRuns = 1, nGPUs = 1;
  int maxIters = 1000, zoom = 27615; int block_size = 32;
  bool output = 1; bool concCpuGpuExec = 0;
  if (argc < 10) {
    if (msl::isRootProcess()) {
      std::cout << std::endl << std::endl;
      std::cout << "Usage: " << argv[0] << " #rows #cols #maxIters #zoom #blockSize #nConcKernels #nRuns #nGPUs concCPUGPUExec" << std::endl;
    }
  } else {
    rows = atoi(argv[1]);
    cols = atoi(argv[2]);
    maxIters = atoi(argv[3]);
    zoom = atoi(argv[4]);
    block_size = atoi(argv[5]);
    msl::setNumConcurrentKernels(atoi(argv[6]));
    nRuns = atoi(argv[7]);
    nGPUs = atoi(argv[8]);
    concCpuGpuExec = (atoi(argv[9]));
    output = false;
  }

  msl::setNumRuns(nRuns);
  msl::setNumGpus(nGPUs);
  msl::setTaskGroupSize(block_size * block_size);
  msl::setFarmStatistics(1);

  const float center_x = -0.73f;
  const float center_y = -0.16f;

	mandelbrotSet = new int*[rows];
	for (int i = 0; i < rows; i++) {
		mandelbrotSet[i] = new int[cols];
	}

	// Functors
	Init init(rows, cols, block_size);
	Iterate iter(maxIters, center_x, center_y, zoom, rows, cols);
	Fin fin;

	// Create a skeleton topology using C++ constructors.
	msl::Initial<int, Init> p1(init);
	msl::Farm<int, Result, Iterate> p3(iter, nWorkers, concCpuGpuExec);
	msl::Final<Result, Fin> p4(fin);
	msl::Pipe p5(p1, p3, p4);

	// start the calculation.
	msl::startTiming();
	for (int run = 0; run < msl::Muesli::num_runs; run++) {
		p5.start();
		MPI_Barrier (MPI_COMM_WORLD);
		if (output) {
      if (msl::Muesli::proc_id == msl::Muesli::num_total_procs - 1) {
        printResult(rows, cols, maxIters);
      }
		}
		reset();
		msl::splitTime(run);
	}
	msl::stopTiming();

	msl::terminateSkeletons();
	return 0;
}

