/*
 * farm_nbody.cpp
 *
 *      Author: Steffen Ernsting <s.ernsting@uni-muenster.de>
 *
 * -------------------------------------------------------------------------------
 *
 * The MIT License
 *
 * Copyright 2014 Steffen Ernsting <s.ernsting@uni-muenster.de>,
 *                Herbert Kuchen <kuchen@uni-muenster.de.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *
 */

#include <iostream>
#include <cstdlib>
#include <fstream>

#include "mpi.h"
#include "muesli.h"
#include "farm.h"
#include "pipe.h"
#include "initial.h"
#include "final.h"
#include "darray.h"
#include "functors.h"
#include "rng.h"
#include "larray.h"

namespace msl {
namespace examples {
namespace farm_nbody {

#define EPSILON         1.0E-10f
#define DT              0.01f

float dt = DT;
int count = 0;

struct Particle
{
public:
	float x, y, z, vx, vy, vz, ax, ay, az, mass, charge;
	int pos;
};

void initParticles(Particle* particles, int num_particles)
{
  Rng rng;
  for (int i = 0; i < num_particles; i++) {
    particles[i] = Particle();
    particles[i].x = rng();
    particles[i].y = rng();
    particles[i].z = rng();
    particles[i].vx = 0.0f;
    particles[i].vy = 0.0f;
    particles[i].vz = 0.0f;
    particles[i].mass = 1.0f;
    particles[i].charge = 1.0f - 2.0f * (float) (i % 2);
    particles[i].pos = i;
  }
}

struct Init
{
	int num_particles;

	Init(int np) : num_particles(np)
	{}

	int* operator() ()
	{
		if (count == num_particles)
			return 0;

		int* res = new int(count);
		count++;
		return res;
	}
};

struct CalcForce : public FarmFunctor<int, Particle>
{
  int tw;
	LArray<Particle> oldParticles;

	CalcForce(DArray<Particle>& p, int tile_width)
		: tw(tile_width), oldParticles(p, Distribution::COPY)
	{
		this->addArgument(&oldParticles);
		this->setTileWidth(tw);
	}

	MSL_USERFUNC
	Particle operator()(int curIndex) const
	{
		Particle curParticle = oldParticles[curIndex];
    curParticle.ax = 0.0f;
    curParticle.ay = 0.0f;
    curParticle.az = 0.0f;

    // calculate forces for the current particle
    for (int t = 0; t < oldParticles.getSize()/tw; t++) {
      auto t_oP = oldParticles.getTile(t, this);

      for (int i = 0; i < tw; i++) {
        // do not evaluate interaction with yourself.
        if (t*tw+i != curIndex) {
          // Evaluate forces that j-particles exert on the i-particle.
          float dx, dy, dz, r2, r, qj_by_r3;

          // Here we absorb the minus sign by changing the order of i and j.
          dx = curParticle.x - t_oP.get(i).x;
          dy = curParticle.y - t_oP.get(i).y;
          dz = curParticle.z - t_oP.get(i).z;

          r2 = dx * dx + dy * dy + dz * dz;
          r = sqrtf(r2);

          // Quench the force if the particles are too close.
          if (r < EPSILON)
            qj_by_r3 = 0.0f;
          else
            qj_by_r3 = t_oP.get(i).charge / (r2 * r);

          // accumulate the contribution from particle j.
          curParticle.ax += qj_by_r3 * dx;
          curParticle.ay += qj_by_r3 * dy;
          curParticle.az += qj_by_r3 * dz;
        }
      }
    }

		return curParticle;
	}
};

struct Fin
{
	DArray<Particle>* particles;

	Fin(DArray<Particle>* da) : particles(da)
	{}

	void operator() (Particle curParticle)
	{
		particles->download();
		Particle* p_vect = particles->getLocalPartition();

		double qidt_by_m, dt_by_2, vx0, vy0, vz0;

		int pos = curParticle.pos;

		// Save old velocities.
		vx0 = curParticle.vx;
		vy0 = curParticle.vy;
		vz0 = curParticle.vz;

		// Now advance the velocity of particle i.
		qidt_by_m = p_vect[pos].charge * dt / p_vect[pos].mass;
		p_vect[pos].vx += curParticle.ax * qidt_by_m;
		p_vect[pos].vy += curParticle.ay * qidt_by_m;
		p_vect[pos].vz += curParticle.az * qidt_by_m;

		// Use average velocity in the interval to advance the particles' positions
		dt_by_2 = 0.5 * dt;
		p_vect[pos].x += (vx0 + p_vect[pos].vx) * dt_by_2;
		p_vect[pos].y += (vy0 + p_vect[pos].vy) * dt_by_2;
		p_vect[pos].z += (vz0 + p_vect[pos].vz) * dt_by_2;
	}
};

void show(DArray<Particle>& da)
{
  Particle* p = da.getLocalPartition();

  for (int i = 0; i < da.getLocalSize(); i++) {
    std::cout << i << ": (" << p[i].x << ", " << p[i].y << ", " << p[i].z << ")" << std::endl;
  }
  std::cout << "--------------------------------------------" << std::endl;
}

}
}
}

int main(int argc, char** argv)
{
  using namespace msl::examples::farm_nbody;

	bool debugCommunication = 0;
	msl::initSkeletons(argc, argv, debugCommunication);

  int nWorkers = 1;
  if (msl::Muesli::num_total_procs < 3) {
    if (msl::isRootProcess()) {
      std::cout << "Start with at least 3 processes!" << std::endl;
    }
    msl::fail_exit();
  } else {
    nWorkers = msl::Muesli::num_total_procs - 2;
  }

	int maxParticles = 64; int time_steps = 5; bool output = 0;
	int tile_width = 8; bool concCpuGpuExec = 0;
	if (argc < 8) {
		if (msl::isRootProcess()) {
			std::cout << "Usage: "
			          << argv[0] << " nParticles time_steps taskGroupSize nGPUs nConcurrentKernels nRuns concCPUGPUExec"
			          << std::endl;
		}
		msl::setTaskGroupSize(16);
		output = 1;
	} else {
		maxParticles = atoi(argv[1]);
		time_steps = atoi(argv[2]);
		msl::setTaskGroupSize(atoi(argv[3]));
    msl::setNumGpus(atoi(argv[4]));
    msl::setNumConcurrentKernels(atoi(argv[5]));
    msl::setNumRuns(atoi(argv[6]));
    concCpuGpuExec = (atoi(argv[7]));
    tile_width = msl::Muesli::threads_per_block;
	}

	msl::setFarmStatistics(1);

	// initialize particle system
	Particle* p = new Particle[maxParticles];
	if (msl::isRootProcess())
		initParticles(p, maxParticles);

	// broadcast particle system
	msl::MSL_Broadcast(0, p, maxParticles);

	// copy distributed array
	msl::DArray<Particle> particles(maxParticles, p, msl::Distribution::COPY);

	// functors
	Init init(maxParticles);
	CalcForce cf(particles, tile_width);
	Fin fin(&particles);

	// create a skeleton topology.
	msl::Initial<int, Init> p1(init);
	msl::Farm<int, Particle, CalcForce> p2(cf, nWorkers, concCpuGpuExec);
	msl::Final<Particle, Fin> p3(fin);
	msl::Pipe p4(p1, p2, p3);

	// timing
	double bcast_start, bcast_time = 0.0;

	// start the computation
	msl::startTiming();
	for (int run = 0; run < msl::Muesli::num_runs; run++) {
		for (int j = 0; j < time_steps; j++) {
			p4.start();
			MPI_Barrier (MPI_COMM_WORLD);
			count = 0;

			bcast_start = MPI_Wtime();
			particles.broadcastPartition(msl::Muesli::num_total_procs - 1);
			bcast_time += (MPI_Wtime() - bcast_start);

			// the last process prints the results.
			if (msl::Muesli::proc_id == msl::Muesli::num_total_procs - 1) {
			  if (output)
			    show(particles);
			}
		}
		msl::splitTime(run);
	}
	msl::stopTiming();

	std::cout << "Broadcast time: " << bcast_time << std::endl;

	msl::terminateSkeletons();

	return 0;
}
