#include <string>
#include <iostream>
#include <fstream>
#include <sstream>
#include <ctime>

#include "farm.h"
#include "pipe.h"
#include "initial.h"
#include "final.h"
#include "raytracer.h"
#include "functors.h"
#include "rng.h"

namespace msl {

namespace examples {

namespace raytracer {

const size_t kNumPixelSamplesU = 4;
const size_t kNumPixelSamplesV = 4;
const size_t kNumLightSamplesU = 4;
const size_t kNumLightSamplesV = 4;

MSL_USERFUNC
Color trace(const Ray& ray, ShapeSet* shapeset)
{
  Rng rng;
  Color result(0.0f, 0.0f, 0.0f);
  Color lightResult;
  Vector toLight;

  Intersection intersection(ray);
  if (!shapeset->intersect(intersection)) {
    return result;
  }

  //Find out what lights the intersected point can see
  Point position = intersection.position();
  for (size_t i = 0; i < shapeset->nLights; i++) {
    // Sample the light
    lightResult = Color(0.0f, 0.0f, 0.0f);
    for (size_t lsv = 0; lsv < kNumLightSamplesV; lsv++) {
      for (size_t lsu = 0; lsu < kNumLightSamplesU; lsu++) {
        Point lightPoint;
        Vector lightNormal;
        RectangleLight* rectLight = &shapeset->lights[i];

        rectLight->sampleSurface(
            (lsu + rng()) / float(kNumLightSamplesU),
            (lsv + rng()) / float(kNumLightSamplesV),
            position, lightPoint, lightNormal);

        // Fire a shadow ray to make sure we can actually see that light position
        toLight = lightPoint - position;
        float lightDistance = toLight.normalize();
        Ray shadowRay(position, toLight, lightDistance);
        Intersection shadowIntersection(shadowRay);
        bool intersected = shapeset->intersect(shadowIntersection);

        if (!intersected) {
          if (intersection.m_rectLight == 0) {
            if (intersection.m_Lambert == 0) {
              lightResult += rectLight->emitted() * intersection.m_colorModifier
                  * intersection.m_Phong->shade(position, intersection.m_normal, ray.m_direction,
                      toLight);
            } else {
              lightResult += rectLight->emitted() * intersection.m_colorModifier
                  * intersection.m_Lambert->shade(position, intersection.m_normal, ray.m_direction,
                      toLight);
            }
          }
        }
        if (shadowIntersection.m_rectLight == rectLight) {
          lightResult += rectLight->emitted() * intersection.m_colorModifier
              * intersection.m_Emitter->shade(position, intersection.m_normal, ray.m_direction, toLight);
        }
      }
    }
    lightResult /= kNumLightSamplesU * kNumLightSamplesV;

    result += lightResult;
  }

  return result;
}

int xCounter = 0, yCounter = 0;

struct Init
{
  int height, width;

  Init(int h, int w)
    : height(h), width(w)
  {}

  Pixel* operator() ()
  {
    if (yCounter == height) {
      return 0;
    }

    Pixel* result = new Pixel(0.0f, 0.0f, 0.0f, xCounter, yCounter);

    if (xCounter == width - 1) {
      yCounter++;
    }
    xCounter = (xCounter + 1) % width;

    return result;
  }

  void reset()
  {
    xCounter = yCounter = 0;
  }
};

struct CastRay : public FarmFunctor<Pixel, Pixel>
{
  LScene scene;
  int height, width;

  CastRay(Scene& s, int h, int w)
    : scene(s), height(h), width(w)
  {
    this->addArgument(&scene);
  }

	MSL_USERFUNC
	Pixel operator()(Pixel p) const
	{
	  // create RNG
    Rng rng;
		// For each sample in the pixel...
		Color pixelColor(0.0f, 0.0f, 0.0f);
		//unsigned int tid = threadIdx.x;

		for (size_t vsi = 0; vsi < kNumPixelSamplesV; ++vsi) {
			for (size_t usi = 0; usi < kNumPixelSamplesU; ++usi) {
				// Calculate a stratified random position within the pixel
				// to hide aliasing.
			  float yu = 1.0f - (p.ypos + (vsi + rng()) / float(kNumPixelSamplesV)) / float(height);
        float xu = (p.xpos + (usi + rng()) / float(kNumPixelSamplesU)) / float(width);

				// Find where this pixel sample hits in the scene
				Ray ray = makeCameraRay(60.0f, Point(0.0f, 5.0f, 15.0f), Point(0.0f, 0.0f, 0.0f),
						Point(0.0f, 1.0f, 0.0f), xu, yu);

				pixelColor += trace(ray, scene.getShapeSet());
			}
		}
		// Divide by the number of pixel samples
		pixelColor /= kNumPixelSamplesU * kNumPixelSamplesV;
		p.setColor(pixelColor);
		return p;
	}
};

struct Fin
{
  DMatrix<Pixel>* img;

  Fin(DMatrix<Pixel>* dm) : img(dm)
  {}

  void operator() (Pixel p)
  {
    img->setLocal(p.xpos, p.ypos, p);
  }
};

} // namespace raytracer
} // namespace examples
} // namespace msl

int main(int argc, char **argv)
{
  using namespace msl::examples::raytracer;

	bool debugCommunication = 0;
	msl::initSkeletons(argc, argv, debugCommunication);

	int nWorkers = 1;
	if (msl::Muesli::num_total_procs < 3) {
	  if (msl::isRootProcess()) {
	    std::cout << "Start with at least 3 processes!" << std::endl;
	  }
	  msl::fail_exit();
	} else {
	  nWorkers = msl::Muesli::num_total_procs - 2;
	}

  int width = 256; int height = 256; bool concCpuGpuExec = 0;
	if (argc < 8) {
		if (msl::isRootProcess()) {
			std::cout << "Usage: " << argv[0] << " imgHeight imgWidth taskGroupSize nGPUs nConcurrentKernels nRuns concCPUGPUExec" << std::endl;
		}
	} else {
	  height = atoi(argv[1]);
	  width = atoi(argv[2]);
	  msl::setTaskGroupSize(atoi(argv[3]));
		msl::setNumGpus(atoi(argv[4]));
		msl::setNumConcurrentKernels(atoi(argv[5]));
		msl::setNumRuns(atoi(argv[6]));
		concCpuGpuExec = (atoi(argv[7]));
		msl::setFarmStatistics(1);
	}

	// The 'scene'
	int nSpheres = 100;
  int nLights = 10;
  Scene scene(nSpheres, nLights);

  // Initialize output image required for the final stage
  msl::DMatrix<Pixel> img(height, width);

  // create functor
  CastRay cr(scene, height, width);
  Init init(height, width);
  Fin fin(&img);

  // Create a skeleton topologie using C++ constructors.
	msl::Initial<Pixel, Init> p1(init);
	msl::Farm<Pixel, Pixel, CastRay> p2(cr, nWorkers, concCpuGpuExec);
	msl::Final<Pixel, Fin> p3(fin);
	msl::Pipe p4(p1, p2, p3);

	// Start the calculation.
	msl::startTiming();
	for (int run = 0; run < msl::Muesli::num_runs; run++) {
		p4.start();
		MPI_Barrier (MPI_COMM_WORLD);
		init.reset();
		msl::splitTime(run);
	}
	msl::stopTiming();

	if (msl::Muesli::proc_id == msl::Muesli::num_total_procs - 1) {
		printRGB(img);
	}

	msl::terminateSkeletons();

	return 0;
}

