/*
 * muesli.cpp
 *
 *      Author: Steffen Ernsting <s.ernsting@uni-muenster.de>
 * 
 * -------------------------------------------------------------------------------
 *
 * The MIT License
 *
 * Copyright 2014 Steffen Ernsting <s.ernsting@uni-muenster.de>,
 *                Herbert Kuchen <kuchen@uni-muenster.de.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *
 */

#include "muesli.h"

int msl::Muesli::proc_id = -1;
int msl::Muesli::proc_entrance;
int msl::Muesli::running_proc_no = 0;
int msl::Muesli::num_total_procs;
int msl::Muesli::num_local_procs; // equals numOfTotalProcs except when nesting DP into TP skeletons
double msl::Muesli::start_time;
char* msl::Muesli::program_name;
int msl::Muesli::distribution_mode;
int msl::Muesli::task_group_size;
int msl::Muesli::num_conc_kernels;
int msl::Muesli::num_threads;
int msl::Muesli::num_runs;
int msl::Muesli::num_gpus;
int msl::Muesli::max_gpus;
int msl::Muesli::threads_per_block;
int msl::Muesli::tpb_x;
int msl::Muesli::tpb_y;
bool msl::Muesli::debug_communication;
bool msl::Muesli::use_timer;
bool msl::Muesli::farm_statistics = 0;
#ifdef __CUDACC__
std::vector<cudaStream_t> msl::Muesli::streams;
#endif
msl::Timer* timer;

void msl::initSkeletons(int argc, char** argv, bool debug_communication)
{
  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &Muesli::num_total_procs);
  MPI_Comm_rank(MPI_COMM_WORLD, &Muesli::proc_id);
  omp_set_nested(1);

  if (1 <= argc) {
    Muesli::program_name = argv[0];
  }

  int device_count = 0;
#ifdef __CUDACC__
  CUDA_CHECK_RETURN(cudaGetDeviceCount(&device_count));
  Muesli::streams.resize(device_count);
  for (size_t i = 0; i < Muesli::streams.size(); i++) {
    cudaSetDevice(i);
    CUDA_CHECK_RETURN(cudaStreamCreate(&Muesli::streams[i]));
  }
#endif
  Muesli::max_gpus = device_count;
  Muesli::num_gpus = device_count;

  Muesli::task_group_size = DEFAULT_TASK_GOUP_SIZE;
  Muesli::num_conc_kernels = DEFAULT_NUM_CONC_KERNELS;
  Muesli::num_threads = omp_get_max_threads();
  Muesli::debug_communication = debug_communication;
  Muesli::num_runs = DEFAULT_NUM_RUNS;
  Muesli::num_local_procs = Muesli::num_total_procs;
  Muesli::proc_entrance = 0;
  setThreadsPerBlock(512); // default for one dimensional thread blocks
  setThreadsPerBlock(16, 16); // default for two dimensional thread blocks
  Muesli::start_time = MPI_Wtime();
}

void msl::terminateSkeletons()
{
  std::ostringstream s;
  std::ostringstream s_time;

  MPI_Barrier(MPI_COMM_WORLD);

  if (Muesli::use_timer) {
    double total_time = timer->totalTime();
    s_time << "Total time: " << total_time << "s" << std::endl;
  }

  if (isRootProcess()) {
    s << std::endl << "Name: " << Muesli::program_name << std::endl;
    s << "Proc: " << Muesli::num_total_procs << std::endl;
#ifdef __CUDACC__
    s << "GPUs per proc: " << Muesli::num_gpus << std::endl;
#elif defined(__ICC) || defined(__INTEL_COMPILER)
    s << "Phi only" << std::endl;
    s << "Threads per proc: " << Muesli::num_threads << std::endl;
#else
    s << "CPU only" << std::endl;
    s << "Threads per proc: " << Muesli::num_threads << std::endl;
#endif
    if (Muesli::use_timer) {
      s << s_time.str();
      delete timer;
    } else {
      s << "Total time: " << MPI_Wtime() - Muesli::start_time << "s" << std::endl;
    }

    printf("%s", s.str().c_str());
  }

#ifdef _CUDACC__
  for (auto it = Muesli::streams.begin(); it != Muesli::streams.end(); ++it) {
    CUDA_CHECK_RETURN(cudaStreamDestroy(*it));
  }
#endif

  MPI_Finalize();
  Muesli::running_proc_no = 0;
}

void msl::printv(const char* format, ...) {
  va_list argp;
  va_start(argp, format);

  if(isRootProcess()) {
    vprintf(format, argp);
  }

  va_end(argp);
}

void msl::setNumThreads(int threads)
{
  Muesli::num_threads = threads;
  omp_set_num_threads(threads);
}

void msl::setNumRuns(int runs)
{
  Muesli::num_runs = runs;
}

void msl::setNumGpus(int num_gpus)
{
  if (num_gpus > 0 && num_gpus <= Muesli::max_gpus)
    Muesli::num_gpus = num_gpus;
}

void msl::setThreadsPerBlock(int tpb)
{
  int threads = tpb;
  if (threads > 1024) {
    std::cout << "Warning: threadsPerBlock > 1024." << std::endl
              << "Setting to 1024." << std::endl;
    threads = 1024;
  }
  Muesli::threads_per_block = threads;
}

void msl::setThreadsPerBlock(int tpbX, int tpbY)
{
  if (tpbX * tpbY > 1024) {
    std::cout << "Warning: threadsPerBlock > 1024." << std::endl
              << "Setting to 32x32." << std::endl;
    Muesli::tpb_x = Muesli::tpb_y = 32;
  }
  Muesli::tpb_x = tpbX;
  Muesli::tpb_y = tpbY;
}

void msl::setNumConcurrentKernels(int num_kernels)
{
  Muesli::num_conc_kernels = num_kernels;
}

void msl::setTaskGroupSize(int size)
{
  Muesli::task_group_size = size;
}

void msl::syncStreams()
{
#ifdef __CUDACC__
  for (int i = 0; i < Muesli::num_gpus; i++) {
    cudaSetDevice(i);
    CUDA_CHECK_RETURN(cudaStreamSynchronize(Muesli::streams[i]));
  }
#endif
}

void msl::startTiming()
{
  Muesli::use_timer = 1;
  MPI_Barrier(MPI_COMM_WORLD);
  timer = new Timer();
}

void msl::splitTime(int run)
{
  double result = timer->splitTime();
  if (isRootProcess()) {
    std::cout << "Run " << run << ": " << result << "s" << std::endl;
  }
}

double msl::stopTiming()
{
  int runs = timer->getNumSplits();
  double total_time = timer->stop();
  msl::printv("\nTime/run: %fs\n", total_time/runs);
  return total_time;
}

bool msl::isRootProcess()
{
  return Muesli::proc_id == 0;
}

void msl::setFarmStatistics(bool val)
{
  Muesli::farm_statistics = val;
}

void msl::fail_exit()
{
  MPI_Barrier(MPI_COMM_WORLD);
  MPI_Finalize();
  exit(EXIT_FAILURE);
}

void msl::throws(const detail::Exception& e)
{
  std::cout << Muesli::proc_id << ": " << e << std::endl;
}
